'''
    AppConfig for django-cleanup, prepare the cache and connect signal handlers
'''
from django.apps import AppConfig

from . import cache, handlers


class CleanupConfig(AppConfig):
    name = 'django_cleanup'
    verbose_name = 'Django Cleanup'
    default = True

    def ready(self):
        cache.prepare(False)
        handlers.connect()

class CleanupSelectedConfig(AppConfig):
    name = 'django_cleanup'
    verbose_name = 'Django Cleanup'

    def ready(self):
        cache.prepare(True)
        handlers.connect()



import logging
import os
import pickle
import re
import tempfile

from django.conf import settings
from django.core.files import File
from django.db import transaction
from django.db.models.fields import NOT_PROVIDED, files

import pytest

from django_cleanup import cache, handlers
from django_cleanup.signals import cleanup_post_delete, cleanup_pre_delete

from . import storage
from .models.app import (
    BranchProduct, Product, ProductIgnore, ProductProxy, ProductUnmanaged, RootProduct)
from .testing_helpers import get_random_pic_name, get_using


LINE = re.compile(r'line \d{1,3}')

TB = '''Traceback (most recent call last):
  File "{handlers}", line xxx, in run_on_commit
    file_.delete(save=False)
  File "{files}", line xxx, in delete
    self.storage.delete(self.name)
  File "{storage}", line xxx, in delete
    os.remove(name)
{error}: [Errno 2] No such file or directory: '{{picture}}\''''



def getTraceback():
    fileabspath = os.path.abspath
    error = 'FileNotFoundError'

    return TB.format(
        handlers=fileabspath(handlers.__file__),
        files=fileabspath(files.__file__),
        storage=fileabspath(storage.__file__),
        error=error)


def _raise(message):
    def _func(x):  # pragma: no cover
        raise Exception(message)
    return _func


@pytest.mark.django_db(transaction=True)
def test_refresh_from_db_without_refresh(picture):
    product = Product.objects.create(image=picture['filename'])
    assert os.path.exists(picture['path'])
    product.refresh_from_db()
    assert id(product.image.instance) == id(product)
    product.image = get_random_pic_name()
    with transaction.atomic(get_using(product)):
        product.save()
    assert not os.path.exists(picture['path'])


@pytest.mark.django_db(transaction=True)
def test_cache_gone(picture):
    product = Product.objects.create(image=picture['filename'])
    assert os.path.exists(picture['path'])
    product.image = get_random_pic_name()
    cache.remove_instance_cache(product)
    with transaction.atomic(get_using(product)):
        product.save()
    assert not os.path.exists(picture['path'])


@pytest.mark.django_db(transaction=True)
def test_storage_gone(picture):
    product = Product.objects.create(image=picture['filename'])
    assert os.path.exists(picture['path'])
    product.image = get_random_pic_name()
    product = pickle.loads(pickle.dumps(product))
    assert hasattr(product.image, 'storage')
    with transaction.atomic(get_using(product)):
        product.save()
    assert not os.path.exists(picture['path'])


@pytest.mark.django_db(transaction=True)
def test_replace_file_with_file(picture):
    product = Product.objects.create(image=picture['filename'])
    assert os.path.exists(picture['path'])
    random_pic_name = get_random_pic_name()
    product.image = random_pic_name
    with transaction.atomic(get_using(product)):
        product.save()
    assert not os.path.exists(picture['path'])
    assert product.image
    new_image_path = os.path.join(settings.MEDIA_ROOT, random_pic_name)
    assert product.image.path == new_image_path


@pytest.mark.django_db(transaction=True)
def test_replace_file_with_blank(picture):
    product = Product.objects.create(image=picture['filename'])
    assert os.path.exists(picture['path'])
    product.image = ''
    with transaction.atomic(get_using(product)):
        product.save()
    assert not os.path.exists(picture['path'])
    assert not product.image
    assert product.image.name == ''


@pytest.mark.django_db(transaction=True)
def test_replace_file_with_none(picture):
    product = Product.objects.create(image=picture['filename'])
    assert os.path.exists(picture['path'])
    product.image = None
    with transaction.atomic(get_using(product)):
        product.save()
    assert not os.path.exists(picture['path'])
    assert not product.image
    assert product.image.name is None


@pytest.mark.django_db(transaction=True)
def test_replace_file_proxy(picture):
    product = ProductProxy.objects.create(image=picture['filename'])
    assert os.path.exists(picture['path'])
    product.image = get_random_pic_name()
    with transaction.atomic(get_using(product)):
        product.save()
    assert not os.path.exists(picture['path'])


@pytest.mark.django_db(transaction=True)
def test_replace_file_unmanaged(picture):
    product = ProductUnmanaged.objects.create(image=picture['filename'])
    assert os.path.exists(picture['path'])
    product.image = get_random_pic_name()
    with transaction.atomic(get_using(product)):
        product.save()
    assert not os.path.exists(picture['path'])


@pytest.mark.django_db(transaction=True)
def test_replace_file_deferred(picture):
    '''probably shouldn't save from a deferred model but someone might do it'''
    product = Product.objects.create(image=picture['filename'])
    assert os.path.exists(picture['path'])
    product_deferred = Product.objects.defer('image_default').get(id=product.id)
    product_deferred.image = get_random_pic_name()
    with transaction.atomic(get_using(product)):
        product_deferred.save()
    assert not os.path.exists(picture['path'])


@pytest.mark.django_db(transaction=True)
def test_remove_model_instance(picture):
    product = Product.objects.create(image=picture['filename'])
    assert os.path.exists(picture['path'])
    with transaction.atomic(get_using(product)):
        product.delete()
    assert not os.path.exists(picture['path'])


@pytest.mark.django_db(transaction=True)
def test_remove_model_instance_default(picture):
    product = Product.objects.create()
    assert product.image_default.path == picture['srcpath']
    assert product.image_default_callable.path == picture['srcpath']
    assert os.path.exists(picture['srcpath'])
    with transaction.atomic(get_using(product)):
        product.delete()
    assert os.path.exists(picture['srcpath'])


@pytest.mark.django_db(transaction=True)
def test_replace_file_with_file_default(picture):
    product = Product.objects.create()
    assert os.path.exists(picture['srcpath'])
    random_pic_name1 = get_random_pic_name()
    random_pic_name2 = get_random_pic_name()
    product.image_default = random_pic_name1
    product.image_default_callable = random_pic_name2
    with transaction.atomic(get_using(product)):
        product.save()
    assert os.path.exists(picture['srcpath'])


@pytest.mark.django_db(transaction=True)
def test_remove_model_instance_ignore(picture):
    product = ProductIgnore.objects.create(image=picture['filename'])
    assert os.path.exists(picture['path'])
    with transaction.atomic(get_using(product)):
        product.delete()
    assert os.path.exists(picture['path'])


@pytest.mark.django_db(transaction=True)
def test_replace_file_with_file_ignore(picture):
    product = ProductIgnore.objects.create(image=picture['filename'])
    assert os.path.exists(picture['path'])
    random_pic_name = get_random_pic_name()
    product.image = random_pic_name
    with transaction.atomic(get_using(product)):
        product.save()
    assert os.path.exists(picture['path'])
    assert product.image
    new_image_path = os.path.join(settings.MEDIA_ROOT, random_pic_name)
    assert product.image.path == new_image_path


@pytest.mark.django_db(transaction=True)
def test_remove_model_instance_proxy(picture):
    product = ProductProxy.objects.create(image=picture['filename'])
    assert os.path.exists(picture['path'])
    with transaction.atomic(get_using(product)):
        product.delete()
    assert not os.path.exists(picture['path'])


@pytest.mark.django_db(transaction=True)
def test_remove_model_instance_unmanaged(picture):
    product = ProductUnmanaged.objects.create(image=picture['filename'])
    assert os.path.exists(picture['path'])
    with transaction.atomic(get_using(product)):
        product.delete()
    assert not os.path.exists(picture['path'])


@pytest.mark.django_db(transaction=True)
def test_remove_model_instance_deferred(picture):
    product = Product.objects.create(image=picture['filename'])
    assert os.path.exists(picture['path'])
    product_deferred = Product.objects.defer('image_default').get(id=product.id)
    with transaction.atomic(get_using(product)):
        product_deferred.delete()
    assert not os.path.exists(picture['path'])


@pytest.mark.django_db(transaction=True)
def test_remove_blank_file(monkeypatch):
    product = Product.objects.create(image='')
    monkeypatch.setattr(
        product.image.storage, 'exists', _raise('should not call exists'))
    monkeypatch.setattr(
        product.image.storage, 'delete', _raise('should not call delete'))
    with transaction.atomic(get_using(product)):
        product.delete()


@pytest.mark.django_db(transaction=True)
def test_remove_not_exists():
    product = Product.objects.create(image='no-such-file')
    with transaction.atomic(get_using(product)):
        product.delete()


@pytest.mark.django_db(transaction=True)
def test_remove_none(monkeypatch):
    product = Product.objects.create(image=None)
    monkeypatch.setattr(
        product.image.storage, 'exists', _raise('should not call exists'))
    monkeypatch.setattr(
        product.image.storage, 'delete', _raise('should not call delete'))
    with transaction.atomic(get_using(product)):
        product.delete()


@pytest.mark.django_db(transaction=True)
def test_exception_on_save(settings, picture, caplog):
    settings.DEFAULT_FILE_STORAGE = 'test.storage.DeleteErrorStorage'
    product = Product.objects.create(image=picture['filename'])
    # simulate a fieldfile that has a storage that raises a filenotfounderror on delete
    assert os.path.exists(picture['path'])
    product.image.delete(save=False)
    product.image = None
    assert not os.path.exists(picture['path'])
    with transaction.atomic(get_using(product)):
        product.save()
    assert not os.path.exists(picture['path'])

    for record in caplog.records:
        assert LINE.sub('line xxx', record.exc_text) == getTraceback().format(picture=picture['path'])
    assert caplog.record_tuples == [
        (
            'django_cleanup.handlers',
            logging.ERROR,
            'There was an exception deleting the file `{}` on field `test.product.image`'.format(
                picture['filename'])
        )
    ]


@pytest.mark.django_db(transaction=True)
def test_cascade_delete(picture):
    root = RootProduct.objects.create()
    branch = BranchProduct.objects.create(root=root, image=picture['filename'])
    assert os.path.exists(picture['path'])
    root = RootProduct.objects.get()
    with transaction.atomic(get_using(root)):
        root.delete()
    assert not os.path.exists(picture['path'])


@pytest.mark.django_db(transaction=True)
def test_file_exists_on_create_and_update():
    # If a filepath is specified which already exists,
    # the FileField generates a random suffix to choose a different location.
    # We need to make sure, that we fetch this change and would delete the correct one
    # on further edits or the final deletion.
    # In this test case it is simulated by using a temporary file located
    # directly within the same directory as the image would be uploaded to.

    upload_to = Product._meta.get_field("image").upload_to
    dst_directory = os.path.join(settings.MEDIA_ROOT, upload_to)
    if not os.path.isdir(dst_directory):
        os.makedirs(dst_directory)

    # create the new product with a file to simulate an "upload"
    # a file aleady exists so the new file is renamed then saved
    with tempfile.NamedTemporaryFile(prefix="f1__", dir=dst_directory) as f1:
        with transaction.atomic():
            product = Product.objects.create(image=File(f1, name=os.path.join(upload_to, os.path.basename(f1.name))))

        assert f1.name != product.image.path
        assert os.path.exists(f1.name)
        assert os.path.exists(product.image.path)

        path_prior_to_edit = product.image.path

        # edit the product to change the product file to a different file
        # check that it deletes the renamed file, not the original existing file
        with tempfile.NamedTemporaryFile(prefix="f2__", dir=dst_directory) as f2:
            with transaction.atomic(get_using(product)):
                product.image = File(f2, name=os.path.join(upload_to, os.path.basename(f2.name)))
                assert f2.name == product.image.path
                product.save()

            assert f1.name != product.image.path
            assert os.path.exists(f1.name)
            assert f2.name != product.image.path
            assert os.path.isfile(f2.name)
            assert os.path.isfile(product.image.path)
            assert not os.path.isfile(path_prior_to_edit)

            with transaction.atomic(get_using(product)):
                product.delete()

            assert os.path.isfile(f1.name)
            assert os.path.isfile(f2.name)
            assert not os.path.isfile(product.image.path)


@pytest.mark.django_db(transaction=True)
def test_signals(picture):
    prekwargs = None
    postkwargs = None
    def assn_prekwargs(**kwargs):
        nonlocal prekwargs
        prekwargs = kwargs

    def assn_postkwargs(**kwargs):
        nonlocal postkwargs
        postkwargs = kwargs
    
    cleanup_pre_delete.connect(assn_prekwargs, dispatch_uid='pre_test_replace_file_with_file_signals')
    cleanup_post_delete.connect(assn_postkwargs, dispatch_uid='post_test_replace_file_with_file_signals')
    product = Product.objects.create(image=picture['filename'])
    random_pic_name = get_random_pic_name()
    product.image = random_pic_name
    with transaction.atomic(get_using(product)):
        product.save()

    assert prekwargs['deleted'] == False
    assert prekwargs['updated'] == True
    assert prekwargs['instance'] == product
    assert prekwargs['file'] is not None
    assert prekwargs['file_name'] == picture['filename']
    assert isinstance(prekwargs['default_file_name'], NOT_PROVIDED) 
    assert prekwargs['model_name'] == 'test.product'
    assert prekwargs['field_name'] == 'image'
    
    assert postkwargs['deleted'] == False
    assert postkwargs['updated'] == True
    assert postkwargs['instance'] == product
    assert postkwargs['file'] is not None
    assert postkwargs['file_name'] == picture['filename']
    assert isinstance(postkwargs['default_file_name'], NOT_PROVIDED) 
    assert postkwargs['model_name'] == 'test.product'
    assert postkwargs['field_name'] == 'image'
    assert postkwargs['success'] == True
    assert postkwargs['error'] is None

    with transaction.atomic(get_using(product)):
        product.delete()

    assert prekwargs['deleted'] == True
    assert prekwargs['updated'] == False
    assert prekwargs['instance'] == product
    assert prekwargs['file'] is not None
    assert prekwargs['file_name'] == random_pic_name
    assert isinstance(prekwargs['default_file_name'], NOT_PROVIDED) 
    assert prekwargs['model_name'] == 'test.product'
    assert prekwargs['field_name'] == 'image'
    
    assert postkwargs['deleted'] == True
    assert postkwargs['updated'] == False
    assert postkwargs['instance'] == product
    assert postkwargs['file'] is not None
    assert postkwargs['file_name'] == random_pic_name
    assert isinstance(postkwargs['default_file_name'], NOT_PROVIDED) 
    assert postkwargs['model_name'] == 'test.product'
    assert postkwargs['field_name'] == 'image'
    assert postkwargs['success'] == True
    assert postkwargs['error'] is None

    cleanup_pre_delete.disconnect(None, dispatch_uid='pre_test_replace_file_with_file_signals')
    cleanup_post_delete.disconnect(None, dispatch_uid='post_test_replace_file_with_file_signals')


#region select config
@pytest.mark.CleanupSelectedConfig
@pytest.mark.django_db(transaction=True)
def test__select_config__replace_file_with_file(picture):
    product = Product.objects.create(image=picture['filename'])
    assert os.path.exists(picture['path'])
    random_pic_name = get_random_pic_name()
    product.image = random_pic_name
    with transaction.atomic(get_using(product)):
        product.save()
    assert not os.path.exists(picture['path'])
    assert product.image
    new_image_path = os.path.join(settings.MEDIA_ROOT, random_pic_name)
    assert product.image.path == new_image_path


@pytest.mark.CleanupSelectedConfig
@pytest.mark.django_db(transaction=True)
def test__select_config__replace_file_with_file_ignore(picture):
    product = ProductIgnore.objects.create(image=picture['filename'])
    assert os.path.exists(picture['path'])
    random_pic_name = get_random_pic_name()
    product.image = random_pic_name
    with transaction.atomic(get_using(product)):
        product.save()
    assert os.path.exists(picture['path'])
    assert product.image
    new_image_path = os.path.join(settings.MEDIA_ROOT, random_pic_name)
    assert product.image.path == new_image_path
#endregion



'''
    django-cleanup sends the following signals
'''
from django.dispatch import Signal


__all__ = ['cleanup_pre_delete', 'cleanup_post_delete']


cleanup_pre_delete = Signal()
'''Called just before a file is deleted. Passes a `file` keyword argument.'''


cleanup_post_delete = Signal()
'''Called just after a file is deleted. Passes a `file` keyword argument.'''



'''
    Signal handlers to manage FileField files.
'''
import logging

from django.db.models.signals import post_delete, post_init, post_save, pre_save
from django.db.transaction import on_commit

from . import cache
from .signals import cleanup_post_delete, cleanup_pre_delete


logger = logging.getLogger(__name__)


class FakeInstance:
    '''A Fake model instance to ensure an instance is not modified'''


def cache_original_post_init(sender, instance, **kwargs):
    '''Post_init on all models with file fields, saves original values'''
    cache.make_cleanup_cache(instance)


def fallback_pre_save(sender, instance, raw, update_fields, using, **kwargs):
    '''Fallback to the database to remake the cleanup cache if there is none'''
    if raw:  # pragma: no cover
        return

    if instance.pk and not cache.has_cache(instance):
        try:
            db_instance = sender.objects.get(pk=instance.pk)
        except sender.DoesNotExist:  # pragma: no cover
            return
        cache.make_cleanup_cache(instance, source=db_instance)


def delete_old_post_save(sender, instance, raw, created, update_fields, using,
                         **kwargs):
    '''Post_save on all models with file fields, deletes old files'''
    if raw:
        return

    if not created:
        for field_name, new_file in cache.fields_for_model_instance(instance):
            if update_fields is None or field_name in update_fields:
                old_file = cache.get_field_attr(instance, field_name)
                if old_file != new_file:
                    delete_file(sender, instance, field_name, old_file, using, 'updated')

    # reset cache
    cache.make_cleanup_cache(instance)


def delete_all_post_delete(sender, instance, using, **kwargs):
    '''Post_delete on all models with file fields, deletes all files'''
    for field_name, file_ in cache.fields_for_model_instance(instance):
        delete_file(sender, instance, field_name, file_, using, 'deleted')


def delete_file(sender, instance, field_name, file_, using, reason):
    '''Deletes a file'''

    if not file_.name:
        return

    # add a fake instance to the file being deleted to avoid
    # any changes to the real instance.
    file_.instance = FakeInstance()

    # pickled filefields lose lots of data, and contrary to how it is
    # documented, the file descriptor does not recover them

    model_name = cache.get_model_name(instance)

    # recover the 'field' if necessary
    if not hasattr(file_, 'field'):
        file_.field = cache.get_field(model_name, field_name)()
        file_.field.name = field_name

    # if our file name is default don't delete
    default = file_.field.default if not callable(file_.field.default) else file_.field.default()

    if file_.name == default:
        return

    # recover the 'storage' if necessary
    if not hasattr(file_, 'storage'):
        file_.storage = cache.get_field_storage(model_name, field_name)()

    event = {
        'deleted': reason == 'deleted',
        'model_name': model_name,
        'field_name': field_name,
        'file_name': file_.name,
        'default_file_name': default,
        'file': file_,
        'instance': instance,
        'updated': reason == 'updated'
    }

    # this will run after a successful commit
    # assuming you are in a transaction and on a database that supports
    # transactions, otherwise it will run immediately
    def run_on_commit():
        cleanup_pre_delete.send(sender=sender, **event)
        success = False
        error = None
        try:
            file_.delete(save=False)
            success = True
        except Exception as ex:
            error = ex
            opts = instance._meta
            logger.exception(
                'There was an exception deleting the file `%s` on field `%s.%s.%s`',
                file_, opts.app_label, opts.model_name, field_name)
        cleanup_post_delete.send(sender=sender, error=error, success=success, **event)

    on_commit(run_on_commit, using)


def connect():
    '''Connect signals to the cleanup models'''
    for model in cache.cleanup_models():
        key = '{{}}_django_cleanup_{}'.format(cache.get_model_name(model))
        post_init.connect(cache_original_post_init, sender=model,
                          dispatch_uid=key.format('post_init'))
        pre_save.connect(fallback_pre_save, sender=model,
                         dispatch_uid=key.format('pre_save'))
        post_save.connect(delete_old_post_save, sender=model,
                          dispatch_uid=key.format('post_save'))
        post_delete.connect(delete_all_post_delete, sender=model,
                            dispatch_uid=key.format('post_delete'))



import random
import string

from django.db import router


def get_using(instance):
    return router.db_for_write(instance.__class__, instance=instance)


def get_random_pic_name(length=20):
    return 'pic{}.jpg'.format(
        ''.join(random.choice(string.ascii_letters) for m in range(length)))



'''
    django-cleanup automatically deletes files for FileField, ImageField, and
    subclasses. It will delete old files when a new file is being save and it
    will delete files on model instance deletion.
'''

__version__ = '8.1.0'



'''Public utilities'''
from .cache import (
    get_mangled_ignore as _get_mangled_ignore, get_mangled_select as _get_mangled_select,
    make_cleanup_cache as _make_cleanup_cache)


__all__ = ['refresh', 'cleanup_ignore', 'cleanup_select']


def refresh(instance):
    '''Refresh the cache for an instance'''
    return _make_cleanup_cache(instance)


def ignore(cls):
    '''Mark a model to ignore for cleanup'''
    setattr(cls, _get_mangled_ignore(cls), None)
    return cls
cleanup_ignore = ignore


def select(cls):
    '''Mark a model to select for cleanup'''
    setattr(cls, _get_mangled_select(cls), None)
    return cls
cleanup_select = select



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  Copyright (c) 2012 Ilya Shalyapin
#
#  django-cleanup is free software under terms of the MIT License.
#

import os
import re
from codecs import open as codecs_open

from setuptools import find_packages, setup


def read(*parts):
    file_path = os.path.join(os.path.dirname(__file__), *parts)
    return codecs_open(file_path, encoding='utf-8').read()


def find_version(*parts):
    version_file = read(*parts)
    version_match = re.search(
        r'''^__version__ = ['"]([^'"]*)['"]''', version_file, re.M)
    if version_match:
        return str(version_match.group(1))
    raise RuntimeError('Unable to find version string.')


setup(
    name='django-cleanup',
    version=find_version('src/django_cleanup', '__init__.py'),
    packages=['django_cleanup'],
    package_dir={'': 'src'},
    include_package_data=True,
    requires=['python (>=3.6)', 'django (>=3.2)'],
    description='Deletes old files.',
    long_description=read('README.rst'),
    long_description_content_type='text/x-rst',
    author='Ilya Shalyapin',
    author_email='ishalyapin@gmail.com',
    url='https://github.com/un1t/django-cleanup',
    download_url='https://github.com/un1t/django-cleanup/tarball/master',
    license='MIT License',
    keywords='django',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 3.2',
        'Framework :: Django :: 4.1',
        'Framework :: Django :: 4.2',
        'Framework :: Django :: 5.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Utilities',
    ],
)



''' Our local cache of filefields, everything is private to this package.'''
from collections import defaultdict

from django.apps import apps
from django.db import models
from django.utils.module_loading import import_string


CACHE_NAME = '_django_cleanup_original_cache'


def fields_default():
    return set()
FIELDS = defaultdict(fields_default)


def fields_dict_default():
    return {}
FIELDS_FIELDS = defaultdict(fields_dict_default)
FIELDS_STORAGE = defaultdict(fields_dict_default)
DOTTED_PATH = '{klass.__module__}.{klass.__qualname__}'


# cache init ##


def prepare(select_mode):
    '''Prepare the cache for all models, non-reentrant'''
    if FIELDS:  # pragma: no cover
        return

    for model in apps.get_models():
        if ignore_model(model, select_mode):
            continue
        name = get_model_name(model)
        if model_has_filefields(name):  # pragma: no cover
            continue
        opts = model._meta
        for field in opts.get_fields():
            if isinstance(field, models.FileField):
                add_field_for_model(name, field.name, field)


def add_field_for_model(model_name, field_name, field):
    '''Centralized function to make all our local caches.'''
    # store models that have filefields and the field names
    FIELDS[model_name].add(field_name)
    # store the dotted path of the field class for each field
    # in case we need to restore it later on
    FIELDS_FIELDS[model_name][field_name] = get_dotted_path(field)
    # also store the dotted path of the storage for the same reason
    FIELDS_STORAGE[model_name][field_name] = get_dotted_path(field.storage)


# generators ##


def get_fields_for_model(model_name, exclude=None):
    '''Get the filefields for a model if it has them'''
    if model_has_filefields(model_name):
        fields = FIELDS[model_name]

        if exclude is not None:
            assert isinstance(exclude, set)
            fields = fields.difference(exclude)

        for field_name in fields:
            yield field_name


def fields_for_model_instance(instance, using=None):
    '''
        Yields (name, descriptor) for each file field given an instance

        Can use the `using` kwarg to change the instance that the `FieldFile`
        will receive.
    '''
    if using is None:
        using = instance
    model_name = get_model_name(instance)

    deferred_fields = instance.get_deferred_fields()

    for field_name in get_fields_for_model(model_name, exclude=deferred_fields):
        fieldfile = getattr(instance, field_name, None)
        yield field_name, fieldfile.__class__(using, fieldfile.field, fieldfile.name)


# restore ##


def get_field(model_name, field_name):
    '''Restore a field from its dotted path'''
    return import_string(FIELDS_FIELDS[model_name][field_name])


def get_field_storage(model_name, field_name):
    '''Restore a storage from its dotted path'''
    return import_string(FIELDS_STORAGE[model_name][field_name])


# utilities ##


def get_dotted_path(object_):
    '''get the dotted path for an object'''
    return DOTTED_PATH.format(klass=object_.__class__)


def get_model_name(model):
    '''returns a unique model name'''
    return '{opt.app_label}.{opt.model_name}'.format(opt=model._meta)


def get_mangled_ignore(model):
    '''returns a mangled attribute name specific to the model for ignore functionality'''
    return '_{opt.model_name}__{opt.app_label}_cleanup_ignore'.format(opt=model._meta)


def get_mangled_select(model):
    '''returns a mangled attribute name specific to the model for select functionality'''
    return '_{opt.model_name}__{opt.app_label}_cleanup_select'.format(opt=model._meta)


# booleans ##


def model_has_filefields(model_name):
    '''Check if a model has filefields'''
    return model_name in FIELDS


def ignore_model(model, select_mode):
    '''Check if a model should be ignored'''
    return (not hasattr(model, get_mangled_select(model))) if select_mode else hasattr(model, get_mangled_ignore(model))


# instance functions ##


def remove_instance_cache(instance):
    '''Remove the cache from an instance'''
    if has_cache(instance):
        delattr(instance, CACHE_NAME)


def make_cleanup_cache(instance, source=None):
    '''
        Make the cleanup cache for an instance.

        Can also change the source of the data with the `source` kwarg.
    '''

    if source is None:
        source = instance
    setattr(instance, CACHE_NAME, dict(
        fields_for_model_instance(source, using=instance)))


def has_cache(instance):
    '''Check if an instance has a cache on it'''
    return hasattr(instance, CACHE_NAME)


def get_field_attr(instance, field_name):
    '''Get a value from the cache on an instance'''
    return getattr(instance, CACHE_NAME)[field_name]


# data sharing ##


def cleanup_models():
    '''Get all the models we have in the FIELDS cache'''
    for model_name in FIELDS:
        yield apps.get_model(model_name)


def cleanup_fields():
    '''Get a copy of the FIELDS cache'''
    return FIELDS.copy()



import os

from django.conf import settings
from django.db import transaction

import pytest

from django_cleanup.signals import cleanup_pre_delete

from .testing_helpers import get_using


@pytest.mark.django_db(transaction=True)
def test_sorlthumbnail_replace(settings, picture):
    # https://github.com/mariocesar/sorl-thumbnail
    models = pytest.importorskip("test.models.integration")
    ProductIntegration = models.ProductIntegration
    sorl_delete = models.sorl_delete
    cleanup_pre_delete.connect(sorl_delete)
    from sorl.thumbnail import get_thumbnail
    product = ProductIntegration.objects.create(sorl_image=picture['filename'])
    assert os.path.exists(picture['path'])
    im = get_thumbnail(
        product.sorl_image, '100x100', crop='center', quality=50)
    thumbnail_path = os.path.join(settings.MEDIA_ROOT, im.name)
    assert os.path.exists(thumbnail_path)
    product.sorl_image = 'new.png'
    with transaction.atomic(get_using(product)):
        product.save()
    assert not os.path.exists(picture['path'])
    assert not os.path.exists(thumbnail_path)
    cleanup_pre_delete.disconnect(sorl_delete)


@pytest.mark.django_db(transaction=True)
def test_sorlthumbnail_delete(picture):
    # https://github.com/mariocesar/sorl-thumbnail
    models = pytest.importorskip("test.models.integration")
    ProductIntegration = models.ProductIntegration
    sorl_delete = models.sorl_delete
    cleanup_pre_delete.connect(sorl_delete)
    from sorl.thumbnail import get_thumbnail
    product = ProductIntegration.objects.create(sorl_image=picture['filename'])
    assert os.path.exists(picture['path'])
    im = get_thumbnail(
        product.sorl_image, '100x100', crop='center', quality=50)
    thumbnail_path = os.path.join(settings.MEDIA_ROOT, im.name)
    assert os.path.exists(thumbnail_path)
    with transaction.atomic(get_using(product)):
        product.delete()
    assert not os.path.exists(picture['path'])
    assert not os.path.exists(thumbnail_path)
    cleanup_pre_delete.disconnect(sorl_delete)


@pytest.mark.django_db(transaction=True)
def test_easythumbnails_replace(picture):
    # https://github.com/SmileyChris/easy-thumbnails
    models = pytest.importorskip("test.models.integration")
    ProductIntegration = models.ProductIntegration
    from easy_thumbnails.files import get_thumbnailer
    product = ProductIntegration.objects.create(easy_image=picture['filename'])
    assert os.path.exists(picture['path'])
    im = get_thumbnailer(product.easy_image).get_thumbnail(
        {'size': (100, 100)})
    thumbnail_path = os.path.join(settings.MEDIA_ROOT, im.name)
    assert os.path.exists(thumbnail_path)
    product.easy_image = 'new.png'
    with transaction.atomic(get_using(product)):
        product.save()
    assert not os.path.exists(picture['path'])
    assert not os.path.exists(thumbnail_path)


@pytest.mark.django_db(transaction=True)
def test_easythumbnails_delete(picture):
    # https://github.com/SmileyChris/easy-thumbnails
    models = pytest.importorskip("test.models.integration")
    ProductIntegration = models.ProductIntegration
    from easy_thumbnails.files import get_thumbnailer
    product = ProductIntegration.objects.create(easy_image=picture['filename'])
    assert os.path.exists(picture['path'])
    im = get_thumbnailer(product.easy_image).get_thumbnail(
        {'size': (100, 100)})
    thumbnail_path = os.path.join(settings.MEDIA_ROOT, im.name)
    assert os.path.exists(thumbnail_path)
    with transaction.atomic(get_using(product)):
        product.delete()
    assert not os.path.exists(picture['path'])
    assert not os.path.exists(thumbnail_path)
